//   Copyright (c)  2009-2010  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/QQ.H"

#include "CoCoA/NumTheory.H"
#include "CoCoA/OpenMath.H"
#include "CoCoA/convert.H"
#include "CoCoA/error.H"

#include <cmath>
using std::floor;
#include <cstddef>
using std::size_t;
#include <iostream>
using std::ostream;
using std::istream;
#include <limits>
using std::numeric_limits;


namespace CoCoA
{
  
//????????  QQ::QQ(MP_RAT to_be_owned); // for efficient return values in operator+ etc.

  QQ::QQ()
  {
    mpq_init(myRep);
  }

//   QQ::QQ(const MachineInteger& n)
//   {
//     mpq_init(myRep);
//     if (IsNegative(n))
//       mpq_set_si(myRep, AsSignedLong(n), 1);
//     else
//       mpq_set_ui(myRep, AsUnsignedLong(n), 1);
//   }

//   QQ::QQ(const ZZ& N)
//   {
//     mpq_init(myRep);
//     mpq_set_z(myRep, mpzref(N));
//   }

  QQ::QQ(const MachineInteger& n1, const MachineInteger& n2, ReduceFlag status)
  {
    if (IsZero(n2))
      CoCoA_ERROR(ERR::DivByZero, "QQ(n1,n2)");
    mpq_init(myRep);
    myAssign(ZZ(n1), ZZ(n2), status);
//    BELOW IS ORIGINAL CODE -- slightly better 'cos does not create temporaries.
//     const bool IsNegativeFraction = IsNegative(n1) ^ IsNegative(n2);
//     mpq_set_ui(myRep, abs(n1), abs(n2));
//     if (status == NotReduced)
//       mpq_canonicalize(myRep);
//     else
//       CoCoA_ASSERT(gcd(n1,n2) == 1);
//     if (IsNegativeFraction)
//       mpq_neg(myRep, myRep);
  }

  QQ::QQ(const MachineInteger& n1, const ZZ& N2, ReduceFlag status)
  {
    if (IsZero(N2))
      CoCoA_ERROR(ERR::DivByZero, "QQ(n1,N2)");
    mpq_init(myRep);
    myAssign(ZZ(n1), N2, status);
  }

  QQ::QQ(const ZZ& N1, const MachineInteger& n2, ReduceFlag status)
  {
    if (IsZero(n2))
      CoCoA_ERROR(ERR::DivByZero, "QQ(N1,n2)");
    mpq_init(myRep);
    myAssign(N1, ZZ(n2), status);
  }

  QQ::QQ(const ZZ& N1, const ZZ& N2, ReduceFlag status)
  {
    if (IsZero(N2))
      CoCoA_ERROR(ERR::DivByZero, "QQ(N1,N2)");
    mpq_init(myRep);
    myAssign(N1, N2, status);
  }


  QQ::QQ(const std::string& str)
  {
    mpq_init(myRep);
    convert(*this, str);
  }

//   QQ::QQ(const char* str)
//   {
//     mpq_init(myRep);
//     // Need the next "if" because QQ(0) comes here!!
//     if (str == 0)
//       CoCoA_ERROR(ERR::BadArg, "QQ(char*) ctor, zero integer or NULL ptr");
//     convert(*this, str);
//   }


  QQ::QQ(const QQ& from)
  {
    mpq_init(myRep);
    mpq_set(myRep, from.myRep);
  }


// QQ& clone() const;

  QQ::~QQ()
  {
    mpq_clear(myRep);
  }


  // NOTE: This is NOT EXCEPTION CLEAN if the GMP fns can throw.
  void QQ::myAssign(const ZZ& N1, const ZZ& N2, ReduceFlag status/*=NotReduced*/)
  {
    CoCoA_ASSERT(!IsZero(N2));
    const bool IsNegativeFraction = (N1 < 0) ^ (N2 < 0);
    mpz_abs(mpq_numref(myRep), mpzref(N1));
    mpz_abs(mpq_denref(myRep), mpzref(N2));
    if (status == NotReduced)
      mpq_canonicalize(myRep);
    else
      CoCoA_ASSERT(gcd(N1,N2) == 1);
    if (IsNegativeFraction)
      mpq_neg(myRep, myRep);
  }


  QQ& QQ::operator=(const QQ& rhs)
  {
    mpq_set(myRep, rhs.myRep);
    return *this;
  }

    // -------- functions that modify at least one argument or `*this' ----------

  QQ& QQ::operator+=(const QQ& rhs)
  {
    mpq_add(myRep, myRep, rhs.myRep);
    return *this;
  }

  QQ& QQ::operator-=(const QQ& rhs)
  {
    mpq_sub(myRep, myRep, rhs.myRep);
    return *this;
  }

  QQ& QQ::operator*=(const QQ& rhs)
  {
//     if (mpz_sgn(mpq_numref(myRep)) == 0) return *this;
//     if (mpz_sgn(mpq_numref(rhs.myRep)) == 0) return operator=(rhs);
    mpq_mul(myRep, myRep, rhs.myRep);
    return *this;
  }

  QQ& QQ::operator/=(const QQ& rhs)
  {
    if (mpz_sgn(mpq_numref(rhs.myRep)) == 0)
      CoCoA_ERROR(ERR::DivByZero, "q1 /= q2");
    mpq_div(myRep, myRep, rhs.myRep);
    return *this;
  }
                        
    // Same but with RHS a ZZ...
  QQ& QQ::operator=(const ZZ& rhs)
  {
    mpq_set_z(myRep, mpzref(rhs));
    return *this;
  }


  // SLUG: impl is needlessly inefficient: makes useless copy in D
  QQ& QQ::operator+=(const ZZ& rhs)
  {
    ZZ D(CopyFromMPZ, mpq_denref(myRep));
    ZZ tmp = rhs*D;
    mpz_add(mpq_numref(myRep), mpq_numref(myRep), mpzref(tmp));
    // no need to call mpq_canonicalize
    return *this;
  }

  // SLUG: impl is needlessly inefficient: makes useless copy in D
  QQ& QQ::operator-=(const ZZ& rhs)
  {
    ZZ D(CopyFromMPZ, mpq_denref(myRep));
    ZZ tmp = rhs*D;
    mpz_sub(mpq_numref(myRep), mpq_numref(myRep), mpzref(tmp));
    // no need to call mpq_canonicalize
    return *this;
  }

  QQ& QQ::operator*=(const ZZ& rhs)
  {
    return operator*=(QQ(rhs,1));
  }

  QQ& QQ::operator/=(const ZZ& rhs)
  {
    if (IsZero(rhs))
      CoCoA_ERROR(ERR::DivByZero, "Q /= N");
    // Could be more efficient if "*this" is 0.
    return operator/=(QQ(rhs,1));
  }
                        
    // Same but with RHS a MachinInteger...
  QQ& QQ::operator= (const MachineInteger& rhs)
  {
    if (IsNegative(rhs))
      mpq_set_si(myRep, AsSignedLong(rhs), 1);
    else
      mpq_set_ui(myRep, AsUnsignedLong(rhs), 1);
    return *this;
  }

  QQ& QQ::operator+=(const MachineInteger& rhs)
  {
    return operator+=(ZZ(rhs));
  }

  QQ& QQ::operator-=(const MachineInteger& rhs)
  {
    return operator-=(ZZ(rhs));
  }

  QQ& QQ::operator*=(const MachineInteger& rhs)
  {
    return operator*=(ZZ(rhs));
  }

  QQ& QQ::operator/=(const MachineInteger& rhs)
  {
    if (IsZero(rhs))
      CoCoA_ERROR(ERR::DivByZero, "Q /= n");
    return operator/=(ZZ(rhs));
  }


  const QQ& QQ::operator++()
  {
    mpz_add(mpq_numref(myRep), mpq_numref(myRep), mpq_denref(myRep)); // no need to reduce
    return *this;
  }

  const QQ& QQ::operator--()
  {
    mpz_sub(mpq_numref(myRep), mpq_numref(myRep), mpq_denref(myRep));
    return *this;
  }

  const QQ QQ::operator++(int) // INEFFICIENT
  {
    QQ ans(*this);
    operator++();
    return ans;
  }

  const QQ QQ::operator--(int) // INEFFICIENT
  {
    QQ ans(*this);
    operator--();
    return ans;
  }



  // I/O FUNCTIONS

  std::ostream& operator<<(std::ostream& out, const QQ& Q)
  {
    out << num(Q);
    if (!IsOneDen(Q))
      out << "/" << den(Q);
    return out;
  }

  std::istream& operator>>(std::istream& in, QQ& Q)
  {
    if (!in.good()) return in;
    ZZ N;
    in >> N;
    if (!in.good()) return in;
    const char slash = in.peek();
    if (slash != '/')
    {
      Q = N;
      return in;
    }
    ZZ D;
    in >> D;
    if (!in.good()) return in;
    QQ ans(N,D);
    swap(Q, ans); // really an assignment
    return in;
  }


  OpenMathOutput& operator<<(OpenMathOutput& OMOut, const QQ& Q)
  {
    OMOut->mySendApplyStart();
    OMOut->mySendSymbol("nums1","rational");
    OMOut->mySend(num(Q));
    OMOut->mySend(den(Q));
    OMOut->mySendApplyEnd();
    return OMOut;
  }


  OpenMathInput& operator>>(OpenMathInput& OMIn, QQ& /*Q*/)
  {
    CoCoA_ERROR(ERR::NYI, "OpenMathInput fn for QQ");
    return OMIn;
  }


  // STANDARD ARITHMETIC OPERATIONS

  void swap(QQ& a, QQ& b)
  {
    mpq_swap(mpqref(a), mpqref(b));
  }

  const ZZ num(const QQ& Q)
  {
    return ZZ(CopyFromMPZ, mpq_numref(mpqref(Q)));
  }

  const ZZ den(const QQ& Q)
  {
    return ZZ(CopyFromMPZ, mpq_denref(mpqref(Q)));
  }

  const QQ abs(const QQ& Q)
  {
    QQ ans;
    mpq_abs(mpqref(ans), mpqref(Q));
    return ans;
  }

  const QQ operator-(const QQ& Q)
  {
    QQ ans;
    mpq_neg(mpqref(ans), mpqref(Q));
    return ans;
  }

  const QQ operator+(const QQ& Q1, const QQ& Q2)
  {
    QQ ans;
    mpq_add(mpqref(ans), mpqref(Q1), mpqref(Q2));
    return ans;
  }

  const QQ operator-(const QQ& Q1, const QQ& Q2)
  {
    QQ ans;
    mpq_sub(mpqref(ans), mpqref(Q1), mpqref(Q2));
    return ans;
  }

  const QQ operator*(const QQ& Q1, const QQ& Q2)
  {
    QQ ans;
    mpq_mul(mpqref(ans), mpqref(Q1), mpqref(Q2));
    return ans;
  }

  const QQ operator/(const QQ& Q1, const QQ& Q2)
  {
    if (IsZero(Q2))
      CoCoA_ERROR(ERR::DivByZero, "Q1/Q2");
    QQ ans;
    mpq_div(mpqref(ans), mpqref(Q1), mpqref(Q2));
    return ans;
  }


  const QQ operator+(const QQ& Q, const ZZ& N)
  {
    QQ ans = Q;
    return ans += N;
//   THE LINES BELOW SHOULD BE MORE EFFICIENT (but they're not very readable).
//     QQ ans;
//     mpz_mul(mpq_denref(ans), mpzref(N), mpq_denref(mpqref(Q)));
//     mpz_add(mpq_numref(mpqref(ans)), mpq_denref(mpqref(ans)), mpq_numref(mpqref(Q)));
//     mpz_set(mpq_denref(mpqref(ans)), mpq_denref(mpqref(Q)));
//     return ans;
  }

  const QQ operator-(const QQ& Q, const ZZ& N)
  {
    QQ ans(Q);
    return ans -= N;
  }

  const QQ operator*(const QQ& Q, const ZZ& N)
  {
    QQ ans(Q);
    return ans *= N;
  }

  const QQ operator/(const QQ& Q, const ZZ& N)
  {
    if (IsZero(N))
      CoCoA_ERROR(ERR::DivByZero, "Q/N");
    QQ ans(Q);
    return ans /= N;
  }

  const QQ operator+(const ZZ& N, const QQ& Q)
  {
    return Q+N;
  }

  const QQ operator-(const ZZ& N, const QQ& Q)
  {
    QQ ans = Q-N;
    return -ans;
  }

  const QQ operator*(const ZZ& N, const QQ& Q)
  {
    return Q*N;
  }

  const QQ operator/(const ZZ& N, const QQ& Q)
  {
    if (IsZero(Q))
      CoCoA_ERROR(ERR::DivByZero, "N/Q");
    return QQ(N,1)/Q;
  }


  const QQ operator+(const QQ& Q, const MachineInteger& n)
  {
    return Q + QQ(n,1);
  }

  const QQ operator-(const QQ& Q, const MachineInteger& n)
  {
    return Q - QQ(n,1);
  }

  const QQ operator*(const QQ& Q, const MachineInteger& n)
  {
    return Q * QQ(n,1);
  }

  const QQ operator/(const QQ& Q, const MachineInteger& n)
  {
    if (IsZero(n))
      CoCoA_ERROR(ERR::DivByZero, "Q/n");
    return Q / QQ(n,1);
  }


  const QQ operator+(const MachineInteger& n, const QQ& Q)
  {
    return QQ(n,1) + Q;
  }

  const QQ operator-(const MachineInteger& n, const QQ& Q)
  {
    return QQ(n,1) - Q;
  }

  const QQ operator*(const MachineInteger& n, const QQ& Q)
  {
    return QQ(n,1) * Q;
  }

  const QQ operator/(const MachineInteger& n, const QQ& Q)
  {
    if (IsZero(Q))
      CoCoA_ERROR(ERR::DivByZero, "n/Q");
    return QQ(n,1) / Q;
  }

  const QQ power(const QQ& base, const ZZ& exponent)
  {
    if (exponent >= 0)
      return QQ(power(num(base), exponent), power(den(base), exponent), QQ::AlreadyReduced);
    if (IsZero(base))
      CoCoA_ERROR(ERR::BadPwrZero, "power(QQ,ZZ)");
    return QQ(power(den(base), -exponent), power(num(base), -exponent), QQ::AlreadyReduced);
  }

  const QQ power(const QQ& base, const MachineInteger& exponent)
  {
    if (!IsNegative(exponent))
      return QQ(power(num(base), exponent), power(den(base), exponent), QQ::AlreadyReduced);
    if (IsZero(base))
      CoCoA_ERROR(ERR::BadPwrZero, "power(QQ,MachineInteger)");
    return QQ(power(den(base), abs(exponent)), power(num(base), abs(exponent)), QQ::AlreadyReduced);
  }



  bool IsZero(const QQ& Q)
  {
    return IsZero(num(Q));
  }


  bool IsOne(const QQ& Q)
  {
    return mpq_cmp_ui(mpqref(Q), 1,1) == 0;
  }


  bool IsMinusOne(const QQ& Q)
  {
    return mpq_cmp_si(mpqref(Q), -1,1) == 0;
  }


  bool IsOneNum(const QQ& Q)
  {
    return mpz_cmp_ui(mpq_numref(mpqref(Q)), 1) == 0;
  }


  bool IsOneDen(const QQ& Q)
  {
    return mpz_cmp_ui(mpq_denref(mpqref(Q)), 1) == 0;
  }


  int sign(const QQ& Q)
  {
    return mpq_sgn(mpqref(Q));
  }


  // COMPARISON FUNCTIONS

  int cmp(const QQ& Q1, const QQ& Q2)
  {
    return mpq_cmp(mpqref(Q1), mpqref(Q2));
  }


  int cmp(const QQ& Q, const ZZ& N)
  {
    return cmp(num(Q), N*den(Q));
  }


  int cmp(const ZZ& N, const QQ& Q)
  {
    return cmp(N*den(Q), num(Q));
  }


  int cmp(const QQ& Q, const MachineInteger& n)
  {
    if (IsNegative(n))
      return mpq_cmp_si(mpqref(Q), AsSignedLong(n),1);
    return mpq_cmp_ui(mpqref(Q), AsUnsignedLong(n),1);
  }


  int cmp(const MachineInteger& n, const QQ& Q)
  {
    return -cmp(Q, n);
  }


  bool operator==(const QQ& Q1, const QQ& Q2)
  {
    return mpq_equal(mpqref(Q1), mpqref(Q2));
  }

  bool operator!=(const QQ& Q1, const QQ& Q2)
  {
    return !(Q1 == Q2);
  }

  bool operator> (const QQ& Q1, const QQ& Q2)
  {
    return cmp(Q1,Q2) > 0;
  }

  bool operator>=(const QQ& Q1, const QQ& Q2)
  {
    return cmp(Q1,Q2) >= 0;
  }

  bool operator< (const QQ& Q1, const QQ& Q2)
  {
    return cmp(Q1,Q2) < 0;
  }

  bool operator<=(const QQ& Q1, const QQ& Q2)
  {
    return cmp(Q1,Q2) <= 0;
  }

                        
  bool operator==(const QQ& Q, const ZZ& N)
  {
    return IsOneDen(Q) && num(Q) == N;
  }

  bool operator!=(const QQ& Q, const ZZ& N)
  {
    return !(Q == N);
  }

  bool operator> (const QQ& Q, const ZZ& N)
  {
    return cmp(Q,N) > 0;
  }

  bool operator>=(const QQ& Q, const ZZ& N)
  {
    return cmp(Q,N) >= 0;
  }

  bool operator< (const QQ& Q, const ZZ& N)
  {
    return cmp(Q,N) < 0;
  }

  bool operator<=(const QQ& Q, const ZZ& N)
  {
    return cmp(Q,N) <= 0;
  }

                        
  bool operator==(const ZZ& N, const QQ& Q)
  {
    return Q == N;
  }

  bool operator!=(const ZZ& N, const QQ& Q)
  {
    return !(Q == N);
  }

  bool operator> (const ZZ& N, const QQ& Q)
  {
    return cmp(N,Q) > 0;
  }

  bool operator>=(const ZZ& N, const QQ& Q)
  {
    return cmp(N,Q) >= 0;
  }

  bool operator< (const ZZ& N, const QQ& Q)
  {
    return cmp(N,Q) < 0;
  }

  bool operator<=(const ZZ& N, const QQ& Q)
  {
    return cmp(N,Q) <= 0;
  }

                        
  bool operator==(const QQ& Q, const MachineInteger& n)
  {
    return IsOneDen(Q) && num(Q) == n;
  }

  bool operator!=(const QQ& Q, const MachineInteger& n)
  {
    return !(Q == n);
  }

  bool operator> (const QQ& Q, const MachineInteger& n)
  {
    return cmp(Q,n) > 0;
  }

  bool operator>=(const QQ& Q, const MachineInteger& n)
  {
    return cmp(Q,n) >= 0;
  }

  bool operator< (const QQ& Q, const MachineInteger& n)
  {
    return cmp(Q,n) < 0;
  }

  bool operator<=(const QQ& Q, const MachineInteger& n)
  {
    return cmp(Q,n) <= 0;
  }

                
  bool operator==(const MachineInteger& n, const QQ& Q)
  {
    return IsOneDen(Q) && num(Q) == n;
  }

  bool operator!=(const MachineInteger& n, const QQ& Q)
  {
    return !(n == Q);
  }

  bool operator> (const MachineInteger& n, const QQ& Q)
  {
    return cmp(n,Q) > 0;
  }

  bool operator>=(const MachineInteger& n, const QQ& Q)
  {
    return cmp(n,Q) >= 0;
  }

  bool operator< (const MachineInteger& n, const QQ& Q)
  {
    return cmp(n,Q) < 0;
  }

  bool operator<=(const MachineInteger& n, const QQ& Q)
  {
    return cmp(n,Q) <= 0;
  }

                        

  // MISCELLANEOUS FUNCTIONS

  double mantissa(const QQ& /*Q*/)
  {
    CoCoA_ERROR(ERR::NYI, "mantissa(QQ)");
    return 0.0;
  }

  long exponent(const QQ& /*Q*/)
  {
    CoCoA_ERROR(ERR::NYI, "exponent(QQ)");
    return 0;
  }

  // BUG BUG BUG Is this log(Q)  or log(abs(Q))???
  double log(const QQ& Q)
  {
    if (IsZero(Q))
      CoCoA_ERROR(ERR::BadArg, "log(Q)");
    return log(num(Q)) - log(den(Q));
  }

  long ILogBase(const QQ& Q, const MachineInteger& base)
  {
    return ILogBase(Q, ZZ(base));
  }

  // BUG: the following code is ugly and inefficient -- someone should improve it!
  // e.g. only rarely should we need to compute pwr.  Can we modify the code so
  // candidate errs in only 1 direction?  Can we merge the two almost identical
  // sections?
  long ILogBase(const QQ& Q, ZZ base)
  {
    base = abs(base);
    if (base < 2) CoCoA_ERROR(ERR::BadArg, "ILogBase: base must be at least 2");
    if (IsZero(Q)) CoCoA_ERROR(ERR::BadArg, "ILogBase: cannot compute log(0)");
    const double ApproxLog = log(Q)/log(base);
    const double delta = 5 * ApproxLog * numeric_limits<double>::epsilon();
///???BUG not yet fully implemented    if (ApproxLog > numeric_limits<long>::max()) CoCoA_ERROR(ERR::ArgTooBig, "ILogBase");
    const long candidate = std::floor(ApproxLog+delta); // probably right but could be off by 1
    if (std::abs(ApproxLog - candidate) > delta)
      return candidate;
    if (candidate >= 0)
    {
      const ZZ pwr = power(base, candidate);
      const int test = cmp(abs(Q), pwr);
      if (test == 0) return candidate;
      if (test < 0) return candidate-1;
      if (abs(Q) >= base*pwr) return candidate+1;
      return candidate;
    }
    // candidate < 0
    const ZZ pwr = power(base, -candidate);
    const QQ shifted = abs(Q)*pwr;
    const int test = cmp(shifted, 1);
    if (test == 0) return candidate;
    if (test < 0) return candidate-1;
    if (shifted >= base) return candidate+1;
    return candidate;
  }


  ZZ floor(const QQ& Q)
  {
    ZZ ans;
    mpz_fdiv_q(mpzref(ans), mpq_numref(mpqref(Q)), mpq_denref(mpqref(Q)));
    return ans;
  }

  ZZ ceil(const QQ& Q)
  {
    ZZ ans;
    mpz_cdiv_q(mpzref(ans), mpq_numref(mpqref(Q)), mpq_denref(mpqref(Q)));
    return ans;
  }


  ZZ round(const QQ& Q)
  {
    if (IsOneDen(Q)) return num(Q);
    return (2*num(Q)+den(Q))/(2*den(Q));
  }


} // end of namespace CoCoA



// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/QQ.C,v 1.13 2011/06/23 16:01:07 abbott Exp $
// $Log: QQ.C,v $
// Revision 1.13  2011/06/23 16:01:07  abbott
// Removed single arg ctor QQ(MachineInteger), & consequential changes.
//
// Revision 1.12  2011/03/01 15:26:10  abbott
// Improved impl of ILogBase -- faster in most cases.
//
// Revision 1.11  2011/02/25 12:06:51  abbott
// Added new fn IsOneNum; also some minor code cleaning in QQ.C
//
// Revision 1.10  2011/01/14 17:23:19  abbott
// Fixed a minor bug in power.
//
// Revision 1.9  2010/12/26 13:03:16  abbott
// Added ILogBase function (to ZZ & QQ).
//
// Revision 1.8  2010/05/07 14:57:52  abbott
// Two main changes:
//   power(QQ,ZZ) now allows negative exponent
//   renamed QQ::AlreadyNormalized to QQ::AlreadyReduced
//           (and allowed denoms to be negative; the ctor then makes them positive).
//
// Revision 1.7  2010/03/22 11:49:28  abbott
// Added ctor from a string.
//
// Revision 1.6  2010/03/18 16:40:42  abbott
// Added missing include directive.
//
// Revision 1.5  2010/03/18 16:34:10  abbott
// Added new pseudo-ctors for QQ with optional flag to indicate that value is already normalized.
// Added OpenMath I/O operators.
//
// Revision 1.4  2009/10/26 15:39:24  bigatti
// -- added CopyFromMPZ in ZZ ctor
//
// Revision 1.3  2009/07/08 12:26:53  abbott
// Added floor and ceil functions for QQs.
// Added example program for QQs.
// Minor correction to convert.C; minor cleaning to ex-ZZ1.C
//
// Revision 1.2  2009/07/06 12:31:26  abbott
// Commented out two unused function arguments (to keep compiler quiet
// when compiling a debugging version).
//
// Revision 1.1  2009/07/02 16:29:42  abbott
// Added new class QQ to represent rational numbers.
// Consequent change to the Makefile.
//
//
