//   Copyright (c)  2010  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/RandomZZStream.H"
#include "CoCoA/error.H"

#include <iostream>
using std::ostream;
#include <limits>
using std::numeric_limits; // only for myGenValue


namespace CoCoA
{

  RandomZZStream::RandomZZStream(const MachineInteger& lwb, const MachineInteger& upb, const MachineInteger& seed):
      myLwb(lwb),
      myUpb(upb),
      myRange(1+myUpb-myLwb),
      myValue(0), // initialized "properly" in the body
      myCounter(0),
      mySeed(abs(seed)),
      myState() // init in body
  {
    if (myLwb > myUpb)
      CoCoA_ERROR(ERR::BadArg, "RandomZZStream: must have lwb <= upb");
    gmp_randinit_mt(myState);
    gmp_randseed_ui(myState, mySeed);
    myGenValue();
  }

  RandomZZStream::RandomZZStream(const MachineInteger& lwb, const ZZ& upb, const MachineInteger& seed):
      myLwb(lwb),
      myUpb(upb),
      myRange(1+myUpb-myLwb),
      myValue(0), // initialized "properly" in the body
      myCounter(0),
      mySeed(abs(seed)),
      myState() // init in body
  {
    if (myLwb > myUpb)
      CoCoA_ERROR(ERR::BadArg, "RandomZZStream: must have lwb <= upb");
    gmp_randinit_mt(myState);
    gmp_randseed_ui(myState, mySeed);
    myGenValue();
  }

  RandomZZStream::RandomZZStream(const ZZ& lwb, const MachineInteger& upb, const MachineInteger& seed):
      myLwb(lwb),
      myUpb(upb),
      myRange(1+myUpb-myLwb),
      myValue(0), // initialized "properly" in the body
      myCounter(0),
      mySeed(abs(seed)),
      myState() // init in body
  {
    if (myLwb > myUpb)
      CoCoA_ERROR(ERR::BadArg, "RandomZZStream: must have lwb <= upb");
    gmp_randinit_mt(myState);
    gmp_randseed_ui(myState, mySeed);
    myGenValue();
  }

  RandomZZStream::RandomZZStream(const ZZ& lwb, const ZZ& upb, const MachineInteger& seed):
      myLwb(lwb),
      myUpb(upb),
      myRange(1+myUpb-myLwb),
      myValue(0), // initialized "properly" in the body
      myCounter(0),
      mySeed(abs(seed)),
      myState() // init in body
  {
    if (myLwb > myUpb)
      CoCoA_ERROR(ERR::BadArg, "RandomZZStream: must have lwb <= upb");
    gmp_randinit_mt(myState);
    gmp_randseed_ui(myState, mySeed);
    myGenValue();
  }


  RandomZZStream::RandomZZStream(const RandomZZStream& copy):
      myLwb(copy.myLwb),
      myUpb(copy.myUpb),
      myRange(copy.myRange),
      myValue(copy.myValue),
      myCounter(copy.myCounter),
      mySeed(copy.mySeed),
      myState() //init in body
  {
    gmp_randinit_set(myState, copy.myState);
  }

  RandomZZStream& RandomZZStream::operator=(const RandomZZStream& rhs)
  {
    if (this == &rhs) return *this; // this is necessary (because of call to gmp_randclear)
    myLwb = rhs.myLwb;
    myUpb = rhs.myUpb;
    myRange = rhs.myRange;
    myValue = rhs.myValue;
    myCounter = rhs.myCounter;
    mySeed = rhs.mySeed;
    gmp_randclear(myState);                 // Together these effect an assignment
    gmp_randinit_set(myState, rhs.myState); //

    return *this;
  }

  RandomZZStream::~RandomZZStream()
  {
    gmp_randclear(myState);
  }


  // Better define this in case some idiot wants to use it...
  // Not inline, so it is as slow as possible >-}
  RandomZZStream RandomZZStream::operator++(int)
  {
    RandomZZStream ans(*this); // Probably not cheap!
    operator++();
    return ans; // Makes another copy upon returning!
  }


  long RandomZZStream::myIndex() const
  {
    return myCounter;
  }


  void RandomZZStream::myGenValue()
  {
    mpz_urandomm(mpzref(myValue), myState, mpzref(myRange));
    myValue += myLwb;
  }


  ostream& operator<<(ostream& out, const RandomZZStream& RLS)
  {
    out << "RandomZZStream(lwb=" << RLS.myLwb
        << ", upb=" << RLS.myUpb
        << ", seed=" << RLS.mySeed
        << ", counter=" << RLS.myCounter << ")";
    return out;
  }

} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/RandomZZStream.C,v 1.3 2011/05/03 09:45:12 abbott Exp $
// $Log: RandomZZStream.C,v $
// Revision 1.3  2011/05/03 09:45:12  abbott
// Changed default seed to 0 (used to be 1).
// Seeding with 0 no longer means "seed from system time".
//
// Revision 1.2  2011/03/14 10:26:21  abbott
// Changed size_t into long.
//
// Revision 1.1  2010/06/29 15:17:02  abbott
// New class RandomZZStream
//
//
